// exvfs.cpp, v1.01 2013/06/21
// coded by asmodean

// contact: 
//   web:   http://asmodean.reverse.net
//   email: asmodean [at] hush.com
//   irc:   asmodean on efnet (irc.efnet.net)

// This tool extracts VF (*.vfs) archives.

#include "as-util.h"

#pragma pack(1)

struct VFHDR1 {
  uint8_t  signature[4]; // "VF\x00\x02"
  uint16_t entry_count;
  uint16_t entry_length;
  uint32_t toc_length;
  uint32_t file_size;
};

struct VFHDR2 {
  uint32_t filenames_length; // in wchar_t
  uint32_t unknown1;
};

struct VFENTRY {
  uint32_t filename_offset; // in wchar_t
  uint16_t unknown1;
  uint32_t unknown2;
  uint32_t offset;
  uint32_t length;
  uint32_t length2;
  uint8_t  unknown3;
};

struct AGFHDR {
  uint8_t  signature[4]; // "AGF"
  uint32_t type;
  uint32_t length;
  uint32_t unknown2; // 0x74 (data offset in type 1)
  uint32_t unknown3; // 0    (data offset in type 2)
  uint32_t unknown4; // 0
  uint32_t unknown5; // 0x5C
  uint32_t width;
  uint32_t height;
  uint32_t width2;
  uint32_t height2;
  uint8_t  unknown6[72];
};

#pragma pack()

bool process_agf(const string&  filename,
                 uint8_t*       buff,
                 uint32_t       len)
{
  auto hdr = (AGFHDR*) buff;

  if (memcmp(hdr->signature, "AGF", 3)) {
    return false;
  }

  uint32_t out_len  = hdr->width * hdr->height * 4;
  auto     out_buff = new uint8_t[out_len];

  auto p = (uint32_t*) (hdr + 1);

  if (hdr->type == 2) {
    p = (uint32_t*) (buff + hdr->unknown3);
  }

  auto out_p   = (uint32_t*) out_buff;
  auto out_end = out_p + out_len / 4;

  while (out_p < out_end) {
    uint32_t flags = *p++;
    uint32_t n     = flags >> 8;
    flags &= 0xFF;

    switch (flags) {
      case 1:
        while (n--) {
          *out_p++ = *p++;
        }
        break;

      case 2:
        while (n--) {
          *out_p++ = *p;
        }
        p++;
        break;

      case 3:
        {
          uint32_t r = n >> 8;
          n &= 0xFF;

          uint32_t* src = NULL;

          while (n--) {
            src = p;

            for (uint32_t i = 0; i < r; i++) {
              *out_p++ = *src++;
            }
          }

          p = src;
        }
        break;

      case 4:
        {
          uint32_t p = n & 0xFFF;
          n >>= 12;

          auto src = out_p - p;

          while (n--) {
            *out_p++ = *src++;
          }
        }
        break;

      default:
        fprintf(stderr, "%s: unrecognized compression code (%d)\n", filename.c_str(), flags);
        delete [] out_buff;
        return false;
    }
  }

  as::write_bmp(as::get_file_prefix(filename) + ".bmp",
                out_buff,
                out_len,
                hdr->width,
                hdr->height,
                4,
                as::WRITE_BMP_FLIP);

  delete [] out_buff;

  return true;
};

int main(int argc, char** argv) {
  if (argc != 2) {
    fprintf(stderr, "exvfs v1.01 by asmodean\n\n");
    fprintf(stderr, "usage: %s <input.vfs>\n", argv[0]);
    return -1;
  }

  string in_filename(argv[1]);

  int fd = as::open_or_die(in_filename, O_RDONLY | O_BINARY);

  VFHDR1 hdr;
  read(fd, &hdr, sizeof(hdr));

  auto entries = new VFENTRY[hdr.entry_count];
  read(fd, entries, sizeof(VFENTRY) * hdr.entry_count);

  VFHDR2 hdr2;
  read(fd, &hdr2, sizeof(hdr2));

  auto filenames = new wchar_t[hdr2.filenames_length];
  read(fd, filenames, sizeof(wchar_t) * hdr2.filenames_length);

  for (uint32_t i = 0; i < hdr.entry_count; i++) {
    string filename = as::convert_wchar(filenames + entries[i].filename_offset);

    uint32_t len  = entries[i].length;
    auto     buff = new uint8_t[len];
    lseek(fd, entries[i].offset, SEEK_SET);
    read(fd, buff, len);

    if (!process_agf(filename, buff, len)) {
      as::write_file(filename, buff, len);
    }

    delete [] buff;
  }

  delete [] filenames;
  delete [] entries;

  close(fd);

  return 0;
}
